/* 
 * Prototype test 001
 * ------------------
 * A running light to check valid port configuration and pinout.
 * Connect JP_LED to the appropriate port header in order to test it.
 *
 * Also tests the external crystal oscillator and the internal PLL to bring
 * the system up to 32MHz. Use an 8MHz main oscillator.
 *
 * test protocol
 * -------------
 * PORTA: works
 * PORTB: not fully tested because of JTAG connection
 * TODO
 *
 * (cl) Michel Gross, 06.2009
 */

#include <avr/io.h>
#include <util/delay.h>

#include "ebi_driver.h"

#define PORT PORTC // specific port to be tested

#define FLASH_SIZE (8 * 1024 * 1024L)
#define FLASH_START 0x000000L


void configure_clock(void);
void configure_sdram(void);

void CCPIOWrite(volatile uint8_t * address, uint8_t value); 
void init1 (void) __attribute__ ((naked)) __attribute__ ((section (".init1")));

void writeSDRAM(uint32_t addr, uint8_t data);
uint8_t readSDRAM(uint32_t addr);
void fillSDRAM(void);

void init1(void)
{
  configure_clock();
  configure_sdram();
}

int main(void)
{
	uint16_t status = 1;
	static bool flashOk = true;
	uint32_t addr;
	uint8_t dat;

	PORT.DIR = 0xff; // all outputs
	PORT.OUT = 0xff; // all off

	/* Fill SDRAM with assembler routine */
//	fillSDRAM();

	/* Fill SDRAM with data. */
	for (addr = 0x4000, dat = 0; addr < FLASH_SIZE; addr++)
	{
		PORT.OUT = ~(uint8_t)(addr >> 16);
		writeSDRAM(addr, dat++);
	}


	/* Read back from SDRAM and verify */
	for (addr = 0x4000, dat = 0; addr < FLASH_SIZE; addr++)
	{
		PORT.OUT = ~((uint8_t)(addr >> 16) | 0x80);
		if (readSDRAM(addr) != dat++)
		{
			flashOk = false;
			break;
		}
	}

	if (!flashOk)
	{
		while(1)
		{
			nop();
		}
	}

	while (1)
	{ // program main loop
		_delay_ms(100);
		PORT.OUT = ~status;

		if ((status <<= 1) == (1 << 8))
		{
			status = 1;
		}
	}

	return 0;
}

/*
 * Sets up the system clock.
 */
void configure_clock(void) {
    //  Enable the external clock and wait for it to be stable
    OSC.XOSCCTRL  = OSC_FRQRANGE_2TO9_gc | OSC_XOSCSEL_XTAL_16KCLK_gc;
    OSC.CTRL     |= OSC_XOSCEN_bm;
    while ((OSC.STATUS & OSC_XOSCRDY_bm) == 0);

    //  Setup and enable the PLL
    OSC.PLLCTRL = OSC_PLLSRC_XOSC_gc | (0x04 << OSC_PLLFAC_gp);
    OSC.CTRL     |= OSC_PLLEN_bm;
    while ((OSC.STATUS & OSC_PLLRDY_bm) == 0);

    //  Setup the clock prescalers
    //  Note this operation is protected by the CCP register
    CCPIOWrite(&CLK.PSCTRL, (CLK_PSADIV_1_gc | CLK_PSBCDIV_1_1_gc));

    //  Switch to the PLL as the clock source
    CCPIOWrite(&CLK.CTRL, CLK_SCLKSEL_PLL_gc);

    //  Lock the setting cown
    //CLK.LOCK = CLK_LOCK_bm; 
}

/*
 * Configure EBI for MT48LC16M4A2-7E SDRAM
 */
void configure_sdram(void) {

	/* Set signals which are active-low to high value */
	PORTH.OUT = 0x0F;

	/* Configure bus pins as outputs(except for data lines). */
	PORTH.DIR = 0xFF;
	PORTK.DIR = 0xFF;
	PORTJ.DIR = 0xF0;

	/* Adpated to the project of michel gross */
	/* Set direction and output value of Chip Select line (F1). */
	PORTF.DIR = 0x02;
	PORTF.OUT = 0x00; 

	/* Initialize EBI. */
	EBI_Enable( EBI_SDDATAW_4BIT_gc,
	            EBI_LPCMODE_ALE1_gc,
	            EBI_SRMODE_ALE12_gc,
	            EBI_IFMODE_3PORT_gc );

	/* Initialize SDRAM */
	EBI_EnableSDRAM( EBI_CS_ASPACE_8MB_gc,   /* 8 MB (16M x 4 bit) address space. */
	                 (void *) FLASH_START,   /* Base address. */
	                 true,                   /* 3 cycle CAS Latency. */
	                 false,                  /* 11 Row bits. */
	                 EBI_SDCOL_10BIT_gc,     /* 10 Column bits. */
	                 EBI_MRDLY_2CLK_gc,      /* 2 cycle Mode Register Delay. */
	                 EBI_ROWCYCDLY_5CLK_gc,  /* 5 cycle Row Cycle Delay. */
	                 EBI_RPDLY_2CLK_gc,      /* 2 cycle Row to Pre-charge Delay. */
	                 EBI_WRDLY_2CLK_gc,      /* 2 cycle Write Recovery Delay. */
	                 EBI_ESRDLY_5CLK_gc,     /* 5 cycle Exit Self Refresh to Active Delay. */
	                 EBI_ROWCOLDLY_2CLK_gc,  /* 2 cycle Row to Column Delay. */
	                 0x03FF,                 /* 1023 cycle Refresh Period (32.8 ms @ 2MHz). */
	                 0x0100 );               /* 256 cycle Initialization Delay (128 us @ 2MHz). */
}

void CCPIOWrite(volatile uint8_t * address, uint8_t value) {

   volatile uint8_t * tmpAddr = address;
   
   asm volatile(
		"movw r30,  %0"	      "\n\t"
		"ldi  r16,  %2"	      "\n\t"
		"out   %3, r16"	      "\n\t"
		"st     Z,  %1"       "\n\t"
		:
		: "r" (tmpAddr), "r" (value), "M" (CCP_IOREG_gc), "i" (&CCP)
		: "r16", "r30", "r31"
		);

}

void writeSDRAM(uint32_t addr, uint8_t data)
{
	volatile uint8_t addr_hi8 = (uint8_t)((addr >> 16) & 0xFF);
	volatile uint16_t addr_lo16 = (uint16_t)(addr & 0xFFFF);

	asm volatile (
		"sts  %3, %1"            "\n\t"
		"st   Z, %2"             "\n\t"
		"sts  %3, __zero_reg__"  "\n\t"
		:
		: "z" (addr_lo16), "r" (addr_hi8), "r" (data), "i" (&CPU_RAMPZ)
		);
}

uint8_t readSDRAM(uint32_t addr)
{
	volatile uint8_t retval;
	volatile uint8_t addr_hi8 = (uint8_t)((addr >> 16) & 0xFF);
	volatile uint16_t addr_lo16 = (uint16_t)(addr & 0xFFFF);

	asm volatile (
		"sts  %3, %2"            "\n\t"
		"ld   %0, Z"             "\n\t"
		"sts  %3, __zero_reg__"  "\n\t"
		: "=r" (retval)
		: "z" (addr_lo16), "r" (addr_hi8), "i" (&CPU_RAMPZ)
		);

	return retval;
}

void fillSDRAM(void)
{
	asm volatile (
		"eor  r16, r16"           "\n\t"
		"eor  r30, r30"           "\n\t"
		"ldi  r31, 0x40"          "\n"

		"1:"                      "\n\t"
		"sts  %0, r16"            "\n\t"

		"st   Z, r30"             "\n\t"
		"adiw r30, 1"             "\n\t"
		"adc  r16, __zero_reg__"  "\n\t"
		"cpi  r16, 0x80"          "\n\t"
		"brlo 1b"                 "\n\t"

		"sts %0, __zero_reg__"    "\n\t"
		:
		: "i" (&CPU_RAMPZ)
		: "r16", "r30", "r31"
	);
}
